/* Sentinel - IRC Statistical and Operator Services
** ircd.c - ircd parsing related functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: ircd.c,v 1.314 2003/12/19 15:11:18 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "numeric.h"
#include "text.h"
#include "event.h"
#include "match.h"
#include "s_string.h"
#include "slog.h"
#include "mem.h"

#include <assert.h>

extern const char *infotext[];
static char errstr[BUFSIZE];

static void HandleCollide();
static void CollideDie();
static void Trace(struct User *);
static void ParseStats(struct User *, char);

#ifdef SUPPORT_MSG_OPERS
static void MsgOpers(char *, char *);
#endif

static int m_time(char *, int, char **);
static int m_version(char *, int, char **);
static int m_motd(char *, int, char **);
static int m_privmsg(char *, int, char **);
static int m_notice(char *, int, char **);
static int m_server(char *, int, char **);
static int m_squit(char *, int, char **);
static int m_quit(char *, int, char **);
#ifdef TRACK_VERSION
static int m_351(char *, int, char **);
#endif
#ifdef TRACK_ADMIN
static int m_257(char *, int, char **);
static int m_258(char *, int, char **);
static int m_259(char *, int, char **);
#endif
static int m_mode(char *, int, char **);
static int m_kill(char *, int, char **);
static int m_away(char *, int, char **);
static int m_nick(char *, int, char **);
static int m_ping(char *, int, char **);
static int m_admin(char *, int, char **);
static int m_lusers(char *, int, char **);
static int m_users(char *, int, char **);
static int m_stats(char *, int, char **);
static int m_trace(char *, int, char **);
static int m_info(char *, int, char **);
static int m_whois(char *, int, char **);
#if defined(SPLITSERV) || defined(PUBSTATS)
static int m_invite(char *, int, char **);
#endif
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
static int m_eob(char *, int, char **);
#endif
static int m_error(char *, int, char **);
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_svsnick(char *, int, char **);
static int m_svsmode(char *, int, char **);
static int m_svskill(char *, int, char **);
#ifndef IRCD_HYBSVS
static int m_burst(char *, int, char **);
#endif
#if defined(IRCD_HYBBG) && defined(OPER_MODES)
static int m_svshint(char *, int, char **);
#endif
#endif
#if defined(SUPPORT_HOSTSERV)
static int m_svshost(char *, int, char **);
static int m_svsident(char *, int, char **);
#endif
static int m_sjoin(char *, int, char **);
static int m_join(char *, int, char **);
static int m_part(char *, int, char **);
static int m_kick(char *, int, char **);
static int m_sping(char *, int, char **);
#ifdef LISTSERV
static int m_topic(char *, int, char **);
static int m_tb(char *, int, char **);
static int m_tburst(char *, int, char **);
#endif

#ifdef IRCD_BAHAMUT
static int m_hybrid_sjoin(char *, int, char **);
static int m_df_sjoin(char *, int, char **);
static int m_cli_sjoin(char *, int, char **);
#endif

static void ircd_initialize(void);
static void add_msg_to_hash_table(char *, struct MsgTable *);
static struct MsgTable *msg_hash[MSG_TABLE_SIZE];

/* Order does not matter now, the command will have to be added in
** capital letters due to a "limitation" of the hash table
*/
static struct MsgTable msg_table[] =
  {
    { "PING", (long) 0, (long) 0, NULL, m_ping },
    { "PRIVMSG", (long) 0, (long) 0, NULL, m_privmsg },
    { "NICK", (long) 0, (long) 0, NULL, m_nick },
    { "MODE", (long) 0, (long) 0, NULL, m_mode },
    { "QUIT", (long) 0, (long) 0, NULL, m_quit },
    { "AWAY", (long) 0, (long) 0, NULL, m_away },
    { "KILL", (long) 0, (long) 0, NULL, m_kill },
    { "SERVER", (long) 0, (long) 0, NULL, m_server },
    { "SQUIT", (long) 0, (long) 0, NULL, m_squit },
#ifdef TRACK_VERSION
    { "351", (long) 0, (long) 0, NULL, m_351 },
#endif
#ifdef TRACK_ADMIN
    { "257", (long) 0, (long) 0, NULL, m_257 },
    { "258", (long) 0, (long) 0, NULL, m_258 },
    { "259", (long) 0, (long) 0, NULL, m_259 },
#endif
    { "NOTICE", (long) 0, (long) 0, NULL, m_notice },
    { "VERSION", (long) 0, (long) 0, NULL, m_version },
    { "MOTD", (long) 0, (long) 0, NULL, m_motd },
    { "ADMIN", (long) 0, (long) 0, NULL, m_admin },
    { "LUSERS", (long) 0, (long) 0, NULL, m_lusers },
    { "USERS", (long) 0, (long) 0, NULL, m_users },
    { "STATS", (long) 0, (long) 0, NULL, m_stats },
#if defined(SPLITSERV) || defined(PUBSTATS)
    { "INVITE", (long) 0, (long) 0, NULL, m_invite },
#endif
    { "WHOIS", (long) 0, (long) 0, NULL, m_whois },
    { "TRACE", (long) 0, (long) 0, NULL, m_trace },
    { "TIME", (long) 0, (long) 0, NULL, m_time },
    { "INFO", (long) 0, (long) 0, NULL, m_info },
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
    { "EOB", (long) 0, (long) 0, NULL, m_eob },
#endif
    { "ERROR", (long) 0, (long) 0, NULL, m_error },
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
    { "SVSNICK", (long) 0, (long) 0, NULL, m_svsnick },
    { "SVSMODE", (long) 0, (long) 0, NULL, m_svsmode },
    { "SVSKILL", (long) 0, (long) 0, NULL, m_svskill },
    /* Hybserv2's FORCENICK command */
    { "FORCENICK", (long) 0, (long) 0, NULL, m_svsnick },
#ifndef IRCD_HYBSVS
    { "BURST", (long) 0, (long) 0, NULL, m_burst },
#endif
#endif
#if defined(IRCD_HYBBG) && defined(IRCD_HYBBG)
    { "SVSHINT", (long) 0, (long) 0, NULL, m_svshint },
#endif
#if defined(SUPPORT_HOSTSERV)
    { "SVSHOST", (long) 0, (long) 0, NULL, m_svshost },
    { "CHGHOST", (long) 0, (long) 0, NULL, m_svshost },
    { "SETHOST", (long) 0, (long) 0, NULL, m_svshost },
    { "SVSIDENT", (long) 0, (long) 0, NULL, m_svsident },
    { "CHGIDENT", (long) 0, (long) 0, NULL, m_svsident },
    { "SETIDENT", (long) 0, (long) 0, NULL, m_svsident },
#endif
    { "SJOIN", (long) 0, (long) 0, NULL, m_sjoin },
    { "JOIN", (long) 0, (long) 0, NULL, m_join },
    { "KICK", (long) 0, (long) 0, NULL, m_kick },
    { "PART", (long) 0, (long) 0, NULL, m_part },
    { "SPING", (long) 0, (long) 0, NULL, m_sping },
#ifdef LISTSERV
    { "TOPIC", (long) 0, (long) 0, NULL, m_topic },
    { "TB", (long) 0, (long) 0, NULL, m_tb },
    { "TBURST", (long) 0, (long) 0, NULL, m_tburst },
#endif
    { (char *)0, (long) 0, (long) 0, NULL, (int (*)()) 0 }
  };

static void add_msg_to_hash_table(char *msg, struct MsgTable *m)
{
  m->hash = HASH((unsigned char *) msg, MSG_TABLE_SIZE);
  m->next = msg_hash[m->hash];
  msg_hash[m->hash] = m;
}

void InitMsgHash(void)
{
  int i;

  for (i = 0; msg_table[i].handler != NULL; i++)
  {
    add_msg_to_hash_table(msg_table[i].name, &msg_table[i]);
  }
}

void parse(char *line)
{
  char *origin = NULL;
  char *pos;
  char *command = NULL;
  char *message = NULL;
  char *parv[MAXPARA];
  static char coreLine[BUFSIZE];
  int parc = 0;
  int i;
  int status = -1;
  struct MsgTable *hashent;

  /* Clear the parv */
  for (i = 0; i < MAXPARA; i++)
    parv[i] = NULL;

  /* Clear out the error string each run through parse() */
  errstr[0] = '\0';

  if (line != NULL)
  {
    /* Sometimes we'll get a blank line with just a \n on it...
    ** catch those here....they'll core us later on if we don't
    */
    if (*line == '\n')
      return ;
    if (*line == '\000')
      return ;

    /* For debugging purposes... */
    strlcpy(coreLine, line, BUFSIZE);

    debug("R: %s", line);
#ifdef IO_LOG

    slog(IO_LOGNO, L_NOTICE, "R: %s", line);
#endif

    /* Find the first space */
    if ((pos = strchr(line, ' ')))
    {
      *pos = '\0';
      pos++;
      /* If it starts with a :...we have a prefix/origin
      ** pull the origin off into origin, and have pos for
      ** the command, message will be the part afterwards
      */
      if (*line == ':')
      {
        origin = line;
        origin++;
        if ((message = strchr(pos, ' ')))
        {
          *message = '\0';
          message++;
          command = pos;
        }
        else
        {
          command = pos;
          message = NULL;
        }
      }
      else
      {
        message = pos;
        command = line;
      }
    }
    /* Ok, all the nasty part is over, now we need to make a
    ** parv out of what's left
    */

    if (message)
    {
      if (*message == ':')
      {
        message++;
        parv[0] = message;
        parc = 1;
      }
      else
      {
        parc = tokenize(message, parv);
      }
    }
    else
    {
      parc = 0;
    }

    /* Now we should have origin (or NULL), command, and a parv
    ** with it's accompanying parc...lets make ABSOLUTELY sure
    */
    if (!command)
    {
      slog(DEFAULT_LOG, L_ERR, "Command not found: %s", coreLine);
      return ;
    }

    /* Take the command through the hash table */
    hashent = msg_hash[HASH((unsigned char *) command,
                            MSG_TABLE_SIZE)];
    while ((hashent != NULL) &&
           (strcasecmp(hashent->name, command) != 0))
      hashent = hashent->next;
    if (hashent != NULL)
    {
      status = hashent->handler(origin, parc, parv);
      hashent->counter++;
      if (status == 1)
        slog(DEFAULT_LOG, L_ERR, "Error with command %s: %s", command,
             errstr ? errstr : "No error message");
    }
  }
}

/* All handlers accept the origin as the first parameter (or NULL if there
** is no prefix.  Next is a parameter count, and a parameter vector
** Each handler returns 0 on success or 1 on failure
*/

static int m_time(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    spy("TIME Request by %s", origin);
    sendto_dcc_flag(CF_COMMANDS, "TIME Request by %s", origin);
    sts(":%s %d %s %s :%s", settings.name, RPL_TIME, origin,
        settings.name, irc_time(time(NULL)));
    return 0;
  }

  strcpy(errstr, "No Origin for TIME");
  return 1;
}

static int m_version(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    spy("VERSION Request by %s", origin);
    sendto_dcc_flag(CF_COMMANDS, "VERSION Request by %s", origin);
    sts(":%s %d %s %s %s :%s -> %s", settings.name,
        RPL_VERSION, origin, stats_version,
        settings.name, VersionFlags(), stats_date);
    return 0;
  }

  strcpy(errstr, "No Origin for VERSION");
  return 1;
}

static int m_whois(char *origin, int parc, char *parv[])
{
  if (parc < 2)
  {
    strcpy(errstr, "No Target for WHOIS");
    return 1;
  }
  if (origin)
  {
    struct User *u = finduser(origin);

    spy("Remote WHOIS Request by %s for %s", origin, parv[1]);
    sendto_dcc_flag(CF_COMMANDS,
                    "Remote WHOIS Request by %s for %s", origin, parv[1]);
    if (!irccasecmp(parv[1], AdminServ.nick))
    {
      /* Here is where to answer the WHOIS query */
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, AdminServ.nick,
          AdminServ.user, AdminServ.host,
          AdminServ.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, AdminServ.nick,
          settings.name, settings.desc);
      /* Always true for AdminServ */
      sts(":%s %d %s %s :is an IRC Operator",
          settings.name, RPL_WHOISOPERATOR,
          u->nick, AdminServ.nick);
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          AdminServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, AdminServ.nick);
      return 0;
    }
#ifdef MONSERV
    if (!irccasecmp(parv[1], MonServ.nick))
    {
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, MonServ.nick,
          MonServ.user, MonServ.host,
          MonServ.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, MonServ.nick,
          settings.name, settings.desc);
      sts(":%s %d %s %s :is an IRC Operator",
          settings.name, RPL_WHOISOPERATOR,
          u->nick, MonServ.nick);
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          MonServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, MonServ.nick);
      return 0;
    }
#endif
#ifdef JUPESERV
    if (!irccasecmp(parv[1], JupeServ.nick))
    {
      /* Here is where to answer the WHOIS query */
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, JupeServ.nick,
          JupeServ.user, JupeServ.host,
          JupeServ.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, JupeServ.nick,
          settings.name, settings.desc);
      /* Always true for JupeServ */
      sts(":%s %d %s %s :is an IRC Operator",
          settings.name, RPL_WHOISOPERATOR,
          u->nick, JupeServ.nick);
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          JupeServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, JupeServ.nick);
      return 0;
    }
#endif
#ifdef LISTSERV
    if (!irccasecmp(parv[1], ListServ.nick))
    {
      /* Here is where to answer the WHOIS query */
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, ListServ.nick,
          ListServ.user, ListServ.host,
          ListServ.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, ListServ.nick,
          settings.name, settings.desc);
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          ListServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, ListServ.nick);
      return 0;
    }
#endif
#ifdef STATSERV
    if (!irccasecmp(parv[1], StatServ.nick))
    {
      /* Here is where to answer the WHOIS query */
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, StatServ.nick,
          StatServ.user, StatServ.host, StatServ.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, StatServ.nick,
          settings.name, settings.desc);
      if (IsStSetOper())
      {
        sts(":%s %d %s %s :is an IRC Operator",
            settings.name, RPL_WHOISOPERATOR,
            u->nick, StatServ.nick);
      }
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          StatServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, StatServ.nick);
      return 0;
    }
#endif
#ifdef SPLITSERV
    if (!irccasecmp(parv[1], SplitServ.nick))
    {
      /* Here is where to answer the WHOIS query */
      sts(":%s %d %s %s %s %s * :%s", settings.name,
          RPL_WHOISUSER, u->nick, SplitServ.nick,
          SplitServ.user, SplitServ.host, sp.desc);
      sts(":%s %d %s %s %s :%s", settings.name,
          RPL_WHOISSERVER, u->nick, SplitServ.nick,
          settings.name, settings.desc);
      sts(":%s %d %s %s %d %ld :seconds idle, signon time",
          settings.name, RPL_WHOISIDLE, u->nick,
          SplitServ.nick, 0, me.t_start);
      sts(":%s %d %s %s :End of WHOIS", settings.name,
          RPL_ENDOFWHOIS, u->nick, SplitServ.nick);
      return 0;
    }
#endif
    /* Don't answer any others */
    return 0;
  }

  strcpy(errstr, "No Origin for WHOIS");
  return 1;
}

static int m_info(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    const char **text = infotext;
    spy("INFO Request by %s", origin);
    sendto_dcc_flag(CF_COMMANDS, "INFO Request by %s", origin);
    while (*text)
    {
      if (**text)
        sts(":%s %d %s :%s", settings.name, RPL_INFO,
            origin, *text);
      else
        sts(":%s %d %s : ", settings.name, RPL_INFO,
            origin);
      text++;
    }
    sts("%s %d %s :End of /INFO list.", settings.name,
        RPL_ENDOFINFO, origin);
    return 0;
  }

  strcpy(errstr, "No Origin for INFO");
  return 1;
}

static int m_motd(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    spy("MOTD Request by %s", origin);
    sendto_dcc_flag(CF_COMMANDS, "MOTD Request by %s", origin);
    sts(":%s %d %s :- %s Message of the Day -", settings.name,
        RPL_MOTDSTART, origin, settings.name);
    sts(":%s %d %s :- Sentinel - IRC Statistical and Operator Services",
        settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- For the %s IRC Network", settings.name,
        RPL_MOTD, origin, settings.netname);
    sts(":%s %d %s :- ", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Sentinel is written by W. Campbell, Sren "
        "Straarup,", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Jeremy Johnston, and Ilya Shtift.  Contributions by",
        settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Pantheon, MeTaLRoCk, primalz, Tom^Cat, MoeBass, "
        "spaceman,", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Shadow, vance, Fear, Jammer, idunno, fl_, "
        "ThinkPad", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Lyta, druid, screwedup, damaged.", settings.name,
        RPL_MOTD, origin);
    sts(":%s %d %s :- ", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Some code Copyright: Jonathan George, "
        "Kai Seidler, ", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- ircd-hybrid Team, Chris Behrens, Fred "
        "Jacobs,", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :- Lee Hardy and Ilya Shtift.", settings.name,
        RPL_MOTD, origin);
    sts(":%s %d %s :- Sentinel is originally based on GeoStats.",
        settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :-", settings.name, RPL_MOTD, origin);
    sts(":%s %d %s :End of /MOTD command.", settings.name,
        RPL_ENDOFMOTD, origin);
    return 0;
  }

  strcpy(errstr, "No Origin for MOTD");
  return 1;
}

static int m_privmsg(char *origin, int parc, char *parv[])
{
  struct User *u;

  if (origin != NULL)
  {
    /* parv[0] is the target of the PRIVMSG
    ** parv[1] is the complete message
    */

    /* If these don't exist, we have problems */
    if (parc < 2)
    {
      strcpy(errstr, "Missing parameter in PRIVMSG");
      return 1;
    }

    u = finduser(origin);

    if (u == NULL)
    {
      snprintf(errstr, BUFSIZE, "Unknown user: %s", origin);
      return 1;
    }

#ifdef STATSERV
    if (!irccasecmp(parv[0], StatServ.nick))
    {
#ifdef DISALLOW_NICK_PRIVMSG
      notice(origin, StatServ.nick, "Please use /msg %s@%s message",
              StatServ.user, settings.name);
      notice(origin, StatServ.nick, "Message not delivered.");
      return 0;
    }
#else
      statserv(u, parv[0], parv[1]);
      return 0;
    }
#endif
#endif /* STATSERV */

    /* This checks for channel CTCP messages */
    if ( is_channel(parv[0]) )
    {
#ifdef STATSERV
      if (parv[1][0] == '\001')
      {
        statserv(u, parv[0], parv[1]);
      }
#endif
      /* Channel messages are ignored...but are NOT errors */
#ifdef SPLITSERV
      /* Unless they are in the Split service's channel */
      if (!irccasecmp(parv[0], sp.join_chan)
#ifndef SPLITSERV_CUSTOM
          && (parv[1][0] == '!')
#endif
         )
      {
        splitserv(u, parv[0], parv[1]);
        return 0;
      }
#endif
#ifdef PUBSTATS
      if (!irccasecmp(parv[0], settings.pub_chan))
      {
        pubstats(u, parv[0], parv[1]);
        return 0;
      }
#endif
      return 0;
    }

#ifdef SPLITSERV
    if (!irccasecmp(parv[0], SplitServ.nick))
    {
      splitserv(u, parv[0], parv[1]);
      return 0;
    }
#endif

    if (!irccasecmp(parv[0], AdminServ.nick))
    {
      adminserv(u, parv[1]);
      return 0;
    }

#ifdef JUPESERV
    if (!irccasecmp(parv[0], JupeServ.nick))
    {
      jupeserv(u, parv[1]);
      return 0;
    }
#endif

#ifdef LISTSERV
    if (!irccasecmp(parv[0], ListServ.nick))
    {
      listserv(u, parv[1]);
      return 0;
    }
#endif

#ifdef MONSERV
    if (!irccasecmp(parv[0], MonServ.nick))
    {
      monserv(u, parv[1]);
      return 0;
    }
#endif

#ifdef SUPPORT_MSG_UH
#ifdef STATSERV
    if (!irccasecmp(parv[0], s_UH))
    {
      statserv(u, parv[0], parv[1]);
      return 0;
    }
#endif
#ifdef SPLITSERV
    if (!irccasecmp(parv[0], sp_UH))
    {
      splitserv(u, parv[0], parv[1]);
      return 0;
    }
#endif
    if (!irccasecmp(parv[0], a_UH))
    {
      adminserv(u, parv[1]);
      return 0;
    }
#ifdef JUPESERV
    if (!irccasecmp(parv[0], j_UH))
    {
      jupeserv(u, parv[1]);
      return 0;
    }
#endif
#ifdef LISTSERV
    if (!irccasecmp(parv[0], l_UH))
    {
      listserv(u, parv[1]);
      return 0;
    }
#endif
#ifdef MONSERV
    if (!irccasecmp(parv[0], m_UH))
    {
      monserv(u, parv[1]);
      return 0;
    }
#endif
#endif
#ifdef SUPPORT_MSG_OPERS
    /* This will allow PRIVMSG opers@server in Hybrid 6 and 7
    ** to work.
    */
    if (!irccasecmp(parv[0], s_OPERS))
    {
      MsgOpers(origin, parv[1]);
      return 0;
    }
#endif

  }

  /* We should NOT get here... The only way a PRIVMSG to
  ** something on the service can get here is we used QUOTE
  ** to create the user.  We will ignore those messages.
  */

  snprintf(errstr, BUFSIZE, "Unknown PRIVMSG target: %s", parv[0]);
  return 1;
}

static int m_notice(char *origin, int parc, char *parv[])
{
  if (parc < 2)
  {
    strcpy(errstr, "Missing parameters in NOTICE");
    return 1;
  }
  /* Check for the NOTICE AUTH :... at the beginning of a connection.
  ** It's normal and is NOT an error. NOTICE AUTH doesn't have an
  ** origin (on hybrid...) so the check for an origin must be below
  ** this check.
  */
  if (!strcasecmp(parv[0], "AUTH"))
    return 0;
  if (!origin)
  {
    strcpy(errstr, "NOTICE without an origin");
    return 1;
  }
#ifdef SUPPORT_MSG_OPERS
  if (!strcasecmp(parv[0], s_OPERS))
  {
    MsgOpers(origin, parv[1]);
    return 0;
  }
#endif
#ifdef TRACK_CTCPVERSION
  if (strncasecmp(parv[1], "\001VERSION", 8) == 0) /* Check just 8 characters */
  {
    struct User *u = finduser(origin);
    char *tptr = parv[1] + 8;
    char *endp;

    if (u->version != NULL) /* They already responded once, keep the first */
      return 0;

    if (*tptr == '\0') /* Blank VERSION reply, probably abusive */
      return 0;
    tptr++;
    if ((*tptr == '\0') || (*tptr == '\001')) /* Also blank */
      return 0;
    endp = strchr(tptr, '\001');
    if (endp != NULL)
      *endp = '\0';
    u->version = sstrdup(tptr);
    memcounts.cversion++;
    memcounts.cversionmem += strlen(u->version) + 1;
#ifdef MON_DRONES

    CheckVersionDrones(u, u->version);
#endif

    return 0;
  }
#endif
  /* Ick, this provides a nasty way to flood logs */
  snprintf(errstr, BUFSIZE, "Unknown NOTICE: %s - %s - %s", origin,
           parv[0], parv[1]);
  return 1;
}

static int m_server(char *origin, int parc, char *parv[])
{
  struct Server *s;

  /* Hybrid based servers:  parv[0] = server name, parv[1] = hops,
  ** parv[2] = description, with or without an origin
  ** IRCnet with an origin: parv[0] = server name, parv[1] = hops,
  ** parv[2] = server token, parv[3] = description, Without an origin
  ** IRCnet acts like the rest.
  */
  if (parc < 3)
  {
    strcpy(errstr, "Missing parameters in SERVER");
    return 1;
  }
  if (origin)
  {
    struct Server *sorig = findserver(origin);
#ifdef SPLITSERV

    time_t now = time(NULL);
    struct SpServer *spl = NULL;
    char *td = NULL;
#endif

    s = AddServer(parv[0], atoi(parv[1]), 0, parv[2], sorig);
#ifdef SPLITSERV

    spl = SpFindSplit(parv[0]);
    if (spl)
    {
      if (!is_delayed)
      {
        td = timediff(now - spl->ts);
        slog(SPLIT_LOG, L_NOTICE, "End of Split: %s to %s: %s", parv[0],
             origin, td);
        if (IsSPReport())
        {
          if (IsSPHide())
          {
            if (!IsSHide(s))
            {
              notice(sp.join_chan, SplitServ.nick,
                      "END OF SPLIT: %s to %s: %s", parv[0],
                      IsSHide(sorig) ? "HIDDEN" : origin, td);
            }
          }
          else
          {
            notice(sp.join_chan, SplitServ.nick,
                    "END OF SPLIT: %s to %s: %s", parv[0], origin, td);
          }
        }
      }
    }
    else
    {
      if (!is_delayed)
      {
        slog(SPLIT_LOG, L_NOTICE, "New Server: %s to %s", parv[0], origin);
        if (IsSPReport())
        {
          if (IsSPHide())
          {
            if (!IsSHide(s))
            {
              notice(sp.join_chan, SplitServ.nick, "CONNECT: %s to %s", parv[0],
                      IsSHide(sorig) ? "HIDDEN" : origin);
            }
          }
          else
          {
            notice(sp.join_chan, SplitServ.nick, "CONNECT: %s to %s", parv[0],
                   origin);
          }
        }
      }
    }
    SpNewServer(parv[0], origin);
    SpDelSplit(parv[0]);
#endif

    return 0;
  }
  /* Now parse SERVER without an origin (our uplink) */

  /* Once we receive the SERVER line, we have connected to our uplink
  ** sucessfully.  Now initialize all clients/channels/etc
  */
  ircd_initialize();

  /* The uplink is always token 1 in 2.10 */
  s = AddServer(parv[0], atoi(parv[1]), 1, parv[2], me.s);
  /* This is our uplink, so set me.uplink */
  me.uplink = s;

#ifdef SPLITSERV

  slog(SPLIT_LOG, L_NOTICE, "New Server: %s", parv[0]);
  if (!is_delayed)
  {
    if (IsSPReport())
    {
      if (IsSPHide())
      {
        if (!IsSHide(s))
        {
          notice(sp.join_chan, SplitServ.nick, "CONNECT: %s to %s", parv[0],
                  IsSHide(me.s) ? "HIDDEN" : me.s->name);
        }
      }
      else
      {
        notice(sp.join_chan, SplitServ.nick, "CONNECT: %s to %s", parv[0],
               me.s->name);
      }
    }
  }
  map_uplink = SpNewServer(parv[0], settings.name);
  SpDelSplit(parv[0]);
#endif

  return 0;
}

static int m_squit(char *origin, int parc, char *parv[])
{
#ifdef JUPESERV
  struct Jupe *j = FindServerJupe(parv[0]);
#endif

  DelServer(parv[0]);

#ifdef JUPESERV

  if (j)
  {
    if (settings.flags & FLAG_JUPE_PERSIST)
    {
      alert("%s: JUPE Reintroduced After SQUIT: %s",
            JupeServ.nick, parv[0]);
      sendto_dcc_flag(CF_JUPE,
                      "%s: JUPE Reintroduced After SQUIT: %s",
                      JupeServ.nick, parv[0]);
      InitServerJupe(j);
    }
    else
    {
      alert("%s: JUPE Removed by SQUIT: %s", JupeServ.nick,
            parv[0]);
      sendto_dcc_flag(CF_JUPE,
                      "%s: JUPE Removed by SQUIT: %s", JupeServ.nick,
                      parv[0]);
      DelServerJupe(parv[0]);
      SaveJupes();
    }
    /* Send a hybrid like remote SQUIT wallops */
    wallops(settings.name, "Remote SQUIT %s from %s (%s)",
            parv[0], origin, parv[1] ? parv[1] : "No Reason");
  }
#endif

  return 0;
}

static int m_quit(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    spy("User %s QUIT: %s", origin,
        (parv[0]) ? parv[0] : "NO QUIT MESSAGE");
    sendto_dcc_flag(CF_USER, "User %s QUIT: %s", origin,
                    (parv[0]) ? parv[0] : "NO QUIT MESSAGE");
    DelUser(origin);
    return 0;
  }

  strcpy(errstr, "QUIT without an origin");
  return 1;
}

#ifdef TRACK_VERSION
static int m_351(char *origin, int parc, char *parv[])
{
  char verstr[BUFSIZE];
  int i;

  /* parv[0] = the target (who sent the VERSION query)
  ** parv[1]..[x] = The parts of the VERSION reply
  */
  struct Server *s = NULL;

  if (origin)
  {
    s = findserver(origin);
    if (s != NULL)
    {
      verstr[0] = '\0';

      for (i = 1; i < parc; i++)
      {
        if (strcasecmp(parv[i], s->name) != 0)
        {
          /* Only append the version info if it's NOT the server's name */
          if (verstr[0] == '\0')
          {
            strlcpy(verstr, parv[i], BUFSIZE);
          }
          else
          {
            strlcat(verstr, " ", BUFSIZE);
            strlcat(verstr, parv[i], BUFSIZE);
          }
        }
      }
      /* strcpy safe here, both are BUFSIZE */
      strcpy(s->version, verstr);
      return 0;
    }
  }

  strcpy(errstr, "351 Numeric without an origin");
  return 1;
}
#endif

#ifdef TRACK_ADMIN
static int m_257(char *origin, int parc, char *parv[])
{
  /* parv[0] = the target (who sent the ADMIN query)
  ** parv[1] = The ADMIN reply
  */
  struct Server *s = NULL;

  if (origin)
  {
    s = findserver(origin);
    if (s)
    {
      strlcpy(s->admin1, parv[1], LINE);
      return 0;
    }
  }

  strcpy(errstr, "257 Numeric without an origin");
  return 1;
}

static int m_258(char *origin, int parc, char *parv[])
{
  /* parv[0] = the target (who sent the ADMIN query)
  ** parv[1] = The ADMIN reply
  */
  struct Server *s = NULL;

  if (origin)
  {
    s = findserver(origin);
    if (s)
    {
      strlcpy(s->admin2, parv[1], LINE);
      return 0;
    }
  }

  strcpy(errstr, "258 Numeric without an origin");
  return 1;
}

static int m_259(char *origin, int parc, char *parv[])
{
  /* parv[0] = the target (who sent the ADMIN query)
  ** parv[1] = The ADMIN reply
  */
  struct Server *s = NULL;

  if (origin)
  {
    s = findserver(origin);
    if (s)
    {
      strlcpy(s->admin3, parv[1], LINE);
      return 0;
    }
  }

  strcpy(errstr, "259 Numeric without an origin");
  return 1;
}
#endif

static int m_mode(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    if (parc < 2)
    {
      strcpy(errstr, "Missing parameters in MODE");
      return 1;
    }
    if ( is_channel(parv[0]) )
      ChannelMode(findchannel(parv[0]), parc - 1, &parv[1]);
    else
      UserMode(finduser(origin), parv[1]);
    return 0;
  }

  strcpy(errstr, "MODE without an origin");
  return 1;
}

#if defined(SPLITSERV) || defined(PUBSTATS)
static int m_invite(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    spy("Invite Request by %s to %s channel %s", origin, parv[0],
        parv[1]);
    sendto_dcc_flag(CF_COMMANDS,
                    "Invite Request by %s to %s channel %s", origin,
                    parv[0], parv[1]);
#ifdef SPLITSERV

    if (!irccasecmp(parv[0], SplitServ.nick) &&
        !irccasecmp(parv[1], sp.join_chan))
    {
      init_spjoin();
    }
#endif
#ifdef PUBSTATS
    if (!irccasecmp(parv[0], StatServ.nick) &&
        !irccasecmp(parv[1], settings.pub_chan))
    {
      init_pubjoin();
    }
#endif

    return 0;
  }

  strcpy(errstr, "INVITE without an origin");
  return 1;
}
#endif

#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
static int m_eob(char *origin, int parc, char *parv[])
{
  if (me.uplink != NULL)
  {
    if (strcasecmp(me.uplink->name, origin) == 0)
    {
      SetEOB();
      me.eob_time = time(NULL);
      if (IsNotifyEOB())
      {
        wallops(settings.name, "Burst from the uplink "
                "completed: %d seconds",
                me.eob_time - me.uplink->connected_since);
      }
      return 0;
    }
    strcpy(errstr, "EOB from a server other than the origin");
    return 1;
  }
  strcpy(errstr, "EOB received without being connected to a server");
  return 1;
}
#endif

static int m_error(char *origin, int parc, char *parv[])
{
  if (parv[0] != NULL)
  {
    slog(DEFAULT_LOG, L_ERR, "ERROR from server: %s", parv[0]);
    return 0;
  }
  strcpy(errstr, "NULL ERROR received from server");
  return 1;
}

static int m_kill(char *origin, int parc, char *parv[])
{
  struct User *u = NULL;
  struct Server *s = NULL;
#ifdef JUPESERV

  struct Jupe *j = FindNickJupe(parv[0]);
#endif

  if (parc < 1)
  {
    strcpy(errstr, "No target in KILL");
    return 1;
  }
  if (origin)
  {
    u = finduser(origin);

    spy("Kill Request by %s for %s path: %s", origin, parv[0],
        (parv[1]) ? parv[1] : "NO PATH/REASON");
    sendto_dcc_flag(CF_KILL, "Kill Request by %s for %s path: %s", origin,
                    parv[0], (parv[1]) ? parv[1] : "NO PATH/REASON");
    if (u)
    {
      u->server->stats->operkills++;
    }
    else
    {
      s = findserver(origin);
      if (s)
      {
        s->stats->serverkills++;
      }
    }

    DelUser(parv[0]);
    if (!irccasecmp(parv[0], AdminServ.nick))
    {
#if defined(SPYMODE) || defined(ALERTMODE) || defined(DRONES_CHANNEL)
      ad_delayed = 1;
      eventAdd("check admin delayed", evh_CheckAdDelayed, NULL, DEFAULT_DELAY);
#endif

      init_admin();
#ifdef SPYMODE

      init_spy_join();
#endif
#ifdef ALERTMODE

      init_alert_join();
#endif
#ifdef DRONES_CHANNEL

      init_drone_join();
#endif

      HandleCollide();
#ifdef MONSERV

    }
    else if (!irccasecmp(parv[0], MonServ.nick))
    {
      init_monserv();
      HandleCollide();
#ifdef TRACK_CTCPVERSION

      mn_delayed = 1;
      eventAdd("check monserv delayed", evh_CheckMnDelayed, NULL,
               VERSION_DELAY);
#endif

#endif
#ifdef STATSERV

    }
    else if (!irccasecmp(parv[0], StatServ.nick))
    {
      init_statserv();
      init_join();
#ifdef PUBSTATS

      init_pubjoin();
#endif

      HandleCollide();
#endif
#ifdef SPLITSERV

    }
    else if (!irccasecmp(parv[0], SplitServ.nick))
    {
      init_splitserv();
      init_spjoin();
      HandleCollide();
#endif
#ifdef JUPESERV

    }
    else if (!irccasecmp(parv[0], JupeServ.nick))
    {
      init_jupeserv();
      HandleCollide();
#endif
#ifdef LISTSERV

    }
    else if (!irccasecmp(parv[0], ListServ.nick))
    {
      init_listserv();
      HandleCollide();
#endif

    }

#ifdef JUPESERV

    if (j)
    {
      if (settings.flags & FLAG_JUPE_PERSIST)
      {
        alert("%s: JUPE Reintroduced After KILL: %s",
              JupeServ.nick, parv[0]);
        sendto_dcc_flag(CF_JUPE,
                        "%s: JUPE Reintroduced After KILL: %s",
                        JupeServ.nick, parv[0]);
        InitNickJupe(j);
        HandleCollide();
      }
      else
      {
        alert("%s: JUPE Removed by KILL: %s", JupeServ.nick,
              parv[0]);
        sendto_dcc_flag(CF_JUPE,
                        "%s: JUPE Removed by KILL: %s", JupeServ.nick,
                        parv[0]);
        DelNickJupe(parv[0]);
        SaveJupes();
        /* Don't need HandleCollide() here, as it's unable to get into a kill
        ** war.
        */
      }
    }
#endif

    return 0;
  }

  strcpy(errstr, "KILL without an origin");
  return 1;
}

static int m_away(char *origin, int parc, char *parv[])
{
  /* Hybrid 5 and 6 do not propagate AWAY.  6 can propagate it with a patch,
  ** and csircd 1.14 propagates AWAY.
  */
  struct User *u = NULL;
  if (origin)
  {
    u = finduser(origin);
    if (u)
    {
      if (IsAway(u))
      {
        netinfo.away--;
        u->flags &= ~UFLAG_ISAWAY;
        if (u->server != NULL)
        {
          u->server->away--;
        }
      }
      else
      {
        u->flags |= UFLAG_ISAWAY;
        IncreaseAway(u->server);
      }
      return 0;
    }
  }

  strcpy(errstr, "AWAY without an origin");
  return 1;
}

static int m_nick(char *origin, int parc, char *parv[])
{
  struct User *u = NULL;
  struct Server *s = NULL;

  if (origin)
  {
    if (parc < 1)
      return 1;
    /* With an origin, it's a nick change
    ** NOTE:  IRCnet puts a : in front of the new nick, while
    ** The others don't.  It doesn't matter with the new parser
    ** as it will properly deal with it.  The old one didn't
    */

    /* Incorrect :(
    ** csircd 1.13 and below prefixes NICK lines, violating TSora,
    ** if they are defined as a hub and Sentinel is the 2nd or later server
    ** to connect and be routed.
    ** But... since we can use parc, we can save some cycles
    */
    if (parc <= 2)
    {
      /* 2 would be the new nick and the TS */
      if ( (u = finduser(origin)) )
      {
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
        /* XXX - TS must be tracked :/ */
        Change_User(u, parv[0], 0);
#else

        Change_User(u, parv[0]);
#endif

        return 0;
      }
      snprintf(errstr, BUFSIZE, "Unknown user: %s", origin);
      return 1;
    }
  }
  /* Without an origin, it's a nick introduction...
  ** parv contents vary for each server type...
  ** Hybrid:  0 = nick, 1 = hops, 2 = TS, 3 = modes, 4 = user
  **          5 = host, 6 = server, 7 = gecos
  ** IRCnet: 0 = nick, 1 = hops, 2 = user, 3 = host, 4 = server token,
  **         5 = modes, 6 = gecos
  ** Bahamut:  0 = nick, 1 = hops, 2 = TS, 3 = modes, 4 = user
  **           5 = host, 6 = server, 7 = servicestamp, 8 = gecos
  */

  if (parc < 7)
  {
    strcpy(errstr, "Missing parameter in NICK introduction");
    return 1;
  }
#ifndef IRCD_BAHAMUT
  if (parc < 8)
  {
    strcpy(errstr, "Missing parameter in NICK introduction");
    return 1;
  }
#endif

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
  s = findserver(parv[6]);
  assert(s != NULL);
  spy("User Connect %s from %s@%s on %s gecos: %s", parv[0], parv[4],
      parv[5], parv[6], parv[parc - 1]);
  sendto_dcc_flag(CF_USER, "User Connect %s from %s@%s on %s gecos: %s",
                  parv[0], parv[4], parv[5], parv[6], parv[parc - 1]);
  u = AddUser(parv[0], parv[4], parv[5], s, parv[parc - 1]);
  UserMode(u, parv[3]);
#elif defined(IRCD_BAHAMUT)

  s = findserver(parv[6]);
  assert(s != NULL);
  spy("User Connect %s from %s@%s on %s gecos: %s", parv[0], parv[4],
      parv[5], parv[6], parv[parc - 1]);
  sendto_dcc_flag(CF_USER, "User Connect %s from %s@%s on %s gecos: %s",
                  parv[0], parv[4], parv[5], parv[6], parv[parc - 1]);
  u = AddUser(parv[0], parv[4], parv[5], s, parv[parc - 1]);
  UserMode(u, parv[3]);
#endif

  return 0;
}

static int m_ping(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    /* It's a remote Lag detecting PING */
    if (parc < 2)
    {
      strcpy(errstr, "Missing parameter in PING");
      return 1;
    }
    /* Lag detecting PING's we reverse the order of the two
    ** parameters
    */
    sts(":%s PONG %s :%s", settings.name, parv[1], parv[0]);
    return 0;
  }
  /* Without an origin, it's our uplink checking us...be sure to
  ** answer or we'll ping timeout...this PING is like a heartbeat 
  ** to the service as well.
  */

  if (parc < 1)
  {
    strcpy(errstr, "Missing parameter in PING");
    return 1;
  }
  sts("PONG :%s", parv[0]);
  return 0;
}

#ifdef IRCD_BAHAMUT
static int m_hybrid_sjoin(char *origin, int parc, char *parv[])
#else
static int m_sjoin(char *origin, int parc, char *parv[])
#endif
{
  struct Channel *c;
  int modec = 1;
  char *modev[3];
  int userc;
  char *userv[SJPARA];
  int i;
  time_t ts;

  if (origin != NULL)
  {
    /* :origin SJOIN ts chan modestr [key or limits] :users */
    modev[0]
    = parv[2];
    if (parc > 4)
    {
      modev[modec] = parv[3];
      modec++;
    }
    if (parc > 5)
    {
      modev[modec] = parv[4];
      modec++;
    }
    c = findchannel(parv[1]);
    ts = atol(parv[0]);
    if (c == NULL)
      c = AddChannel(parv[1], ts);
    if (ts < c->ts)
      HandleTSChange(c, ts);
    ChannelMode(c, modec, modev);
    userc = sjtoken(parv[parc - 1], userv);
    for (i = 0; i < userc; i++)
      AddUserToChannel(c, userv[i]);

    return 0;
  }

  strcpy(errstr, "SJOIN without an origin");
  return 1;
}

#ifdef IRCD_BAHAMUT
static int m_sjoin(char *origin, int parc, char *parv[])
{
  if (parc == 2)
    return m_cli_sjoin(origin, parc, parv);
  if (isdigit(parv[1][0]))
    return m_df_sjoin(origin, parc, parv);
  return m_hybrid_sjoin(origin, parc, parv);
}

static int m_df_sjoin(char *origin, int parc, char *parv[])
{
  struct Channel *c;
  int modec = 1;
  char *modev[3];
  int userc;
  char *userv[SJPARA];
  int i;
  time_t ts;

  if (origin != NULL)
  {
    /* :origin SJOIN ts svsts chan modestr [key or limits] :users */
    modev[0] = parv[3];
    ts = atol(parv[0]);
    if (parc > 5)
    {
      modev[modec] = parv[4];
      modec++;
    }
    if (parc > 6)
    {
      modev[modec] = parv[5];
      modec++;
    }
    c = findchannel(parv[2]);
    if (c == NULL)
      c = AddChannel(parv[2], atol(parv[0]));
    if (ts < c->ts)
      HandleTSChange(c, ts);
    ChannelMode(c, modec, modev);
    userc = sjtoken(parv[parc - 1], userv);
    for (i = 0; i < userc; i++)
      AddUserToChannel(c, userv[i]);

    return 0;
  }

  strcpy(errstr, "SJOIN without an origin");
  return 1;
}

static int m_cli_sjoin(char *origin, int parc, char *parv[])
{
  /* Parameter checking is done in the wrapper function */
  struct Channel *c;

  c = findchannel(parv[1]);
  if (c == NULL)
    c = AddChannel(parv[1], atol(parv[0]));
  AddUserToChannel(c, origin);

  return 0;
}
#endif

#ifdef LISTSERV
static int m_topic(char *origin, int parc, char *parv[])
{
  struct Channel *c;

  if (parc < 2)
  {
    strcpy(errstr, "Missing parameter in TOPIC");
    return 1;
  }

  c = findchannel(parv[0]);

#ifdef IRCD_BAHAMUT
  if (parc == 4)
  {
    time_t newts;

    newts = atol(parv[2]);

    /* Logic taken straight from Bahamut's m_topic() */
    if (c->topic != NULL && c->topicts >= newts)
      return 0;

    if (c->topicauthor != NULL)
    {
      memcounts.topicinfo--;
      memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
      free(c->topicauthor);
    }
    c->topicauthor = sstrdup(parv[1]);
    memcounts.topicinfo++;
    memcounts.topicinfomem += strlen(c->topicauthor) + 1;
    c->topicts = newts;
    if (c->topic != NULL)
    {
      memcounts.topic--;
      memcounts.topicmem -= strlen(c->topic) + 1;
      free(c->topic);
    }
    c->topic = sstrdup(parv[3]);
    memcounts.topic++;
    memcounts.topicmem += strlen(c->topic) + 1;
  }
#else
  if (parc == 4)
  {
    if (c->topicauthor != NULL)
    {
      memcounts.topicinfo--;
      memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
      free(c->topicauthor);
    }
    c->topicauthor = sstrdup(parv[1]);
    memcounts.topicinfo++;
    memcounts.topicinfomem += strlen(c->topicauthor);
    c->topicts = atol(parv[2]);
    if (c->topic != NULL)
    {
      memcounts.topic--;
      memcounts.topicmem -= strlen(c->topic) + 1;
      free(c->topic);
    }
    c->topic = sstrdup(parv[3]);
    memcounts.topic++;
    memcounts.topicmem += strlen(c->topic) + 1;
  }
#endif
  else
  {
    if (c->topicauthor != NULL)
    {
      memcounts.topicinfo--;
      memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
      free(c->topicauthor);
    }
    c->topicauthor = sstrdup(origin);
    memcounts.topicinfo++;
    memcounts.topicinfomem += strlen(c->topicauthor) + 1;
    c->topicts = time(NULL);
    if (c->topic != NULL)
    {
      memcounts.topic--;
      memcounts.topicmem -= strlen(c->topic) + 1;
      free(c->topic);
    }
    c->topic = sstrdup(parv[1]);
    memcounts.topic++;
    memcounts.topicmem += strlen(c->topic) + 1;
  }

  c->last = time(NULL);

  return 0;
}

static int m_tb(char *origin, int parc, char *parv[])
{
  /* :sourceserver TB channel topicts [topicsetter] :topic */
  struct Channel *c;
  time_t newts;
  char *tmp;

  if (parc < 3)
  {
    strcpy(errstr, "Missing parameter in TB");
    return 1;
  }

  c = findchannel(parv[0]);

  if (parc == 4)
  {
    newts = atol(parv[1]);

    if (c->topic == NULL || ((c->topicts > 0) && (c->topicts > newts)))
    {
      /* Accept if our topic is NULL OR the passed in topic is older */
      if (c->topicauthor != NULL)
      {
        memcounts.topicinfo--;
        memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
        free(c->topicauthor);
      }
      /* Just get the nick. */
      tmp = strchr(parv[2], '!');
      if (tmp != NULL)
        *tmp = '\0';
      c->topicauthor = sstrdup(parv[2]);
      memcounts.topicinfo++;
      memcounts.topicinfomem += strlen(c->topicauthor) + 1;
      c->topicts = newts;
      if (c->topic != NULL)
      {
        memcounts.topic--;
        memcounts.topicmem -= strlen(c->topic) + 1;
        free(c->topic);
      }
      c->topic = sstrdup(parv[3]);
      memcounts.topic++;
      memcounts.topicmem += strlen(c->topic) + 1;
    }
  }
  else
  {
    newts = atol(parv[1]);

    if (c->topic == NULL || ((c->topicts > 0) && (c->topicts > newts)))
    {
      if (c->topicauthor != NULL)
      {
        memcounts.topicinfo--;
        memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
        free(c->topicauthor);
      }
      c->topicauthor = sstrdup("*unknown*");
      memcounts.topicinfo++;
      memcounts.topicinfomem += 10; /* length of "*unknown*" plus \0 */
      c->topicts = newts;
      if (c->topic != NULL)
      {
        memcounts.topic--;
        memcounts.topicmem -= strlen(c->topic) + 1;
        free(c->topic);
      }
      c->topic = sstrdup(parv[2]);
      memcounts.topic++;
      memcounts.topicmem += strlen(c->topic) + 1;
    }
  }

  c->last = time(NULL);

  return 0;
}

static int m_tburst(char *origin, int parc, char *parv[])
{
  /* :sourceserver TBURST chants chan topicts setter :topic */
  struct Channel *c;
  time_t newtts;
  time_t newcts;
  char *tmp;

  if (parc < 5)
  {
    strcpy(errstr, "Missing parameter in TB");
    return 1;
  }

  c = findchannel(parv[1]);
  newtts = atol(parv[2]);
  newcts = atol(parv[0]);

  /* Only set if the TS of the channel is older or the same as what TBURST
  ** passed us.
  */
  if (c->ts > newcts)
    return 0;

  /* Accept if we don't have a topic OR if the topic we have is newer than
  ** the topic passed (oldest topic is what is accepted)
  */
  if (c->topic == NULL || ((c->topicts > 0) && (c->topicts > newtts)))
  {
    if (c->topicauthor != NULL)
    {
      memcounts.topicinfo--;
      memcounts.topicinfomem -= strlen(c->topicauthor) + 1;
      free(c->topicauthor);
    }
    /* Just get the nick. */
    tmp = strchr(parv[3], '!');
    if (tmp != NULL)
      *tmp = '\0';
    c->topicauthor = sstrdup(parv[3]);
    memcounts.topicinfo++;
    memcounts.topicinfomem += strlen(c->topicauthor) + 1;
    c->topicts = newtts;
    if (c->topic != NULL)
    {
      memcounts.topic--;
      memcounts.topicmem -= strlen(c->topic) + 1;
      free(c->topic);
    }
    c->topic = sstrdup(parv[4]);
    memcounts.topic++;
    memcounts.topicmem += strlen(c->topic);
  }

  c->last = time(NULL);

  return 0;
}
#endif

static int m_join(char *origin, int parc, char *parv[])
{
  struct Channel *c;
  struct User *u;
  dlink_node *dl;
  dlink_node *tdl;
  struct ChanUser *cu;

  int modifiers = 0;

  if (origin != NULL)
  {
    /* JOIN 0 is still always sent as :user JOIN 0 */
    if (parv[0][0] == '0')
    {
      u = finduser(origin);
      /* Remove the user from each channel */
      DLINK_FOREACH_SAFE(dl, tdl, u->channels.head)
      {
        cu = (struct ChanUser *) dl->data;
        DelUserFromChannel(cu->chan, u);
        /* The ChanUser is free'd as a side effect of
        ** DelUserFromChannel()
        */
        dlink_delete(dl, &u->channels);
        dlink_free(dl);
      }
      return 0;
    }
    c = findchannel(parv[0]);
    if (c == NULL)
      c = AddChannel(parv[0], 0); /* TS0 channel */

    AddUserFlagToChannel(c, finduser(origin), modifiers);
    return 0;
  }

  strcpy(errstr, "JOIN without an origin");
  return 1;
}

static int m_part(char *origin, int parc, char *parv[])
{
  struct User *u;
  struct Channel *c;

  if (origin != NULL)
  {
    if (parc < 1)
    {
      strcpy(errstr, "Not enough parameters in PART");
      return 1;
    }

    u = finduser(origin);
    c = findchannel(parv[0]);

    if (u == NULL)
    {
      snprintf(errstr, BUFSIZE, "Unknown user %s", origin);
      return 1;
    }
    if (c == NULL)
    {
      snprintf(errstr, BUFSIZE, "Unknown channel %s",
               parv[0]);
      return 1;
    }

    UnlinkChannelFromUser(c, u);
    DelUserFromChannel(c, u);
    return 0;
  }

  strcpy(errstr, "No origin in PART");
  return 1;
}

static int m_kick(char *origin, int parc, char *parv[])
{
  struct User *u;
  struct Channel *c;

  if (origin != NULL)
  {
    if (parc < 2)
    {
      strcpy(errstr, "Not enough parameters in KICK");
      return 1;
    }

    u = finduser(parv[1]);
    c = findchannel(parv[0]);

    if (u == NULL)
    {
      snprintf(errstr, BUFSIZE, "Unknown user %s", parv[1]);
      return 1;
    }
    if (c == NULL)
    {
      snprintf(errstr, BUFSIZE, "Unknown channel %s",
               parv[0]);
      return 1;
    }

    UnlinkChannelFromUser(c, u);
    DelUserFromChannel(c, u);

#ifdef ENABLE_TAKEOVER

    if (u == AdminServ.u)
    {
      dlink_node *dl;
      DLINK_FOREACH(dl, takeover_list.head)
      {
        if (irccasecmp(c->name, (char *) dl->data) == 0)
        {
          /* This was a channel we have taken over, rejoin */
#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
          sts(":%s SJOIN %ld %s 0 :@%s", settings.name, c->ts, c->name,
              AdminServ.nick);
#endif

          return 0;
        }
      }
    }
#endif
    return 0;
  }

  strcpy(errstr, "No origin in KICK");
  return 1;
}

static int m_sping(char *origin, int parc, char *parv[])
{
  if (origin != NULL)
  {
    if (parc < 4)
    {
      strcpy(errstr, "Not enough parameters in SPING");
      return 1;
    }
    if (match(origin, settings.name) == 0)
    {
      sts(":%s SPONG %s %s %s %s", settings.name, origin, parv[1], parv[2],
          parv[3]);
      return 0;
    }
    /* Addressed to one of my jupes, just ignore */
    return 0;
  }
  strcpy(errstr, "No origin in SPING");
  return 1;
}

static int m_admin(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    /* Is it directed to us or a jupe behind us? */
    if (parv[0])
    {
      spy("ADMIN Request by %s to %s", origin, parv[0]);
      sendto_dcc_flag(CF_COMMANDS,
                      "ADMIN Request by %s to %s", origin,
                      parv[0]);
      /* We should NEVER get an ADMIN query that doesn't
      ** have a valid parameter
      */
      if (!strcasecmp(settings.name, parv[0]))
      {
        sts(":%s %d %s :Administrative info about %s",
            settings.name, RPL_ADMINME, origin,
            settings.name);
        sts(":%s %d %s :%s Network Statistical Services",
            settings.name, RPL_ADMINLOC1, origin,
            settings.netname);
        sts(":%s %d %s :%s", settings.name, RPL_ADMINLOC2,
            origin, settings.admin_name);
        sts(":%s %d %s :%s", settings.name, RPL_ADMINEMAIL,
            origin, settings.admin_email);
        return 0;
      }

      snprintf(errstr, BUFSIZE,
               "Received ADMIN request for unknown/juped leaf: %s",
               parv[0]);
      return 1;
    }
    strcpy(errstr, "Missing parameter in ADMIN");
    return 1;
  }
  strcpy(errstr, "ADMIN without an origin");
  return 1;
}

static int m_users(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    /* Is it directed to us or a jupe behind us? */
    if (parv[0])
    {
      spy("USERS Request by %s to %s", origin, parv[0]);
      sendto_dcc_flag(CF_COMMANDS, "USERS Request by %s to %s", origin,
                      parv[0]);
      if (strcasecmp(settings.name, parv[0]) == 0)
      {
        int count = 1; /* AdminServ */
#ifdef SPLITSERV

        count++;
#endif
#ifdef STATSERV

        count++;
#endif
#ifdef JUPESERV

        count++;
#endif
#ifdef LISTSERV

        count++;
#endif
#ifdef MONSERV

        count++;
#endif

        sts(":%s %d %s :Current local  users: %d Max: %d",
            settings.name, RPL_LOCALUSERS, origin, count,
            count);
        sts(":%s %d %s :Current global users: %ld Max: %ld",
            settings.name, RPL_GLOBALUSERS, origin,
            netinfo.users, netinfo.maxusers);
        return 0;
      }
      snprintf(errstr, BUFSIZE, "Received USERS request for unknown leaf: %s",
               parv[0]);
      return 1;
    }
    strcpy(errstr, "Missing parameter in USERS");
    return 1;
  }
  strcpy(errstr, "USERS without an origin");
  return 1;
}

static int m_lusers(char *origin, int parc, char *parv[])
{
  if (origin)
  {
    /* Is it directed to us or a jupe behind us? */
    if (parv[1])
    {
      spy("LUSERS Request by %s to %s for %s", origin,
          parv[1], parv[0]);
      sendto_dcc_flag(CF_COMMANDS,
                      "LUSERS Request by %s to %s for %s", origin,
                      parv[1], parv[0]);
      /* We should NEVER get an LUSERS query that doesn't
      ** have a valid parameter
      */
      /* Ignore the mask option...it's not totally correct
      ** but why bother
      */
      if (!strcasecmp(settings.name, parv[1]))
      {
        int count = 1; /* AdminServ */
#ifdef SPLITSERV

        count++;
#endif
#ifdef STATSERV

        count++;
#endif
#ifdef JUPESERV

        count++;
#endif
#ifdef LISTSERV

        count++;
#endif
#ifdef MONSERV

        count++;
#endif

        sts(":%s %d %s :There are %ld users and %ld invisible"
            " on %d servers", settings.name,
            RPL_LUSERCLIENT, origin, netinfo.users - netinfo.invis,
            netinfo.invis, netinfo.servers);
        sts(":%s %d %s %d :IRC Operators online", settings.name, RPL_LUSEROP,
            origin, netinfo.opers);
        sts(":%s %d %s %d :channels formed", settings.name, RPL_LUSERCHANNELS,
            origin, netinfo.channels);
        sts(":%s %d %s :I have %d clients and 1 servers", settings.name,
            RPL_LUSERME, origin, count);
        /* I need to send these as some statistical clients
        ** parse these
        */
        sts(":%s %d %s :Current local  users: %d Max: %d",
            settings.name, RPL_LOCALUSERS, origin, count,
            count);
        sts(":%s %d %s :Current global users: %ld Max: %ld",
            settings.name, RPL_GLOBALUSERS, origin,
            netinfo.users, netinfo.maxusers);
        return 0;
      }

      snprintf(errstr, BUFSIZE,
               "Received LUSERS request for unknown leaf: %s",
               parv[1]);
      return 1;
    }
    strcpy(errstr, "Missing parameter in LUSERS");
    return 1;
  }
  strcpy(errstr, "LUSERS without an origin");
  return 1;
}

static int m_stats(char *origin, int parc, char *parv[])
{
  /* parv[0] = type, parv[1] = target */
  struct User *u = NULL;
  if (parc < 2)
  {
    strcpy(errstr, "Missing parameter in STATS");
    return 1;
  }
  if (origin)
  {
    u = finduser(origin);
    spy("STATS %c Request by %s to %s", parv[0][0], origin,
        parv[1]);
    sendto_dcc_flag(CF_COMMANDS, "STATS %c Request by %s to %s", parv[0][0],
                    origin, parv[1]);
    if (u)
    {
      if (!strcasecmp(settings.name, parv[1]))
      {
        ParseStats(u, parv[0][0]);
        return 0;
      }
    }
    snprintf(errstr, BUFSIZE, "Unknown user: %s in STATS", origin);
    return 1;
  }
  strcpy(errstr, "No origin in STATS");
  return 1;
}

static int m_trace(char *origin, int parc, char *parv[])
{
  struct User *u = NULL;
  if (parc < 1)
  {
    strcpy(errstr, "Missing parameter in TRACE");
    return 1;
  }
  if (origin)
  {
    u = finduser(origin);
    spy("TRACE Request by %s for %s", origin, parv[0]);
    sendto_dcc_flag(CF_COMMANDS, "TRACE Request by %s for %s", origin, parv[0]);
    if (u)
    {
      if (!strcasecmp(settings.name, parv[0]))
      {
        Trace(u);
        return 0;
      }
    }
    snprintf(errstr, BUFSIZE, "Unknown user: %s in TRACE", origin);
    return 1;
  }
  strcpy(errstr, "No origin in TRACE");
  return 1;
}

#if defined(IRCD_HYBBG) && defined(OPER_MODES)
static int m_svshint(char *origin, int parc, char *parv[])
{
  struct User *u;
  int i;
  int value;

  if (parc == 0)
  {
    /* No parameters, clears out the list of services admins
    ** and identified users (which we don't care about).
    ** This is horribly inefficient, but I know no other way.
    */
    for (i = 0; i < U_TABLE_SIZE; i++)
    {
      DLINK_FOREACH(dl, userhash[i].head)
      {
        u = dl->data;

        if (u->operflags & OPER_SVSADMIN)
        {
          /* I know it costs as much to check as it
          ** does to simply check, I need to limit the
          ** messages.
          */
          u->operflags &= ~OPER_SVSADMIN;
          alert("Services Admin Status Removal "
                "%s on %s", u->nick,
                u->server->name);
          sendto_dcc_flag(CF_OPER,
                          "Services Admin Status Removal "
                          "%s on %s", u->nick,
                          u->server->name);
        }
      }
    }
    return 0;
  }

  /* Otherwise, it's setting a flag on a user. */
  if (parc < 2)
  {
    strcpy(errstr, "Not enough parameters to SVSHINT");
    return 1;
  }

  value = atoi(parv[0]);
  u = finduser(parv[1]);

  if (value & BG_SVSADMIN)
  {
    u->operflags |= OPER_SVSADMIN;
    alert("New Services Admin %s on %s", u->nick, u->server->name);
    sendto_dcc_flag(CF_OPER, "New Services Admin %s on %s",
                    u->nick, u->server->name);
  }
  /* Ignore services identification */
  return 0;
}
#endif

#if defined(SUPPORT_HOSTSERV)
static int m_svshost(char *origin, int parc, char *parv[])
{
  struct User *u;

  if (origin)
  {
    if (parc < 2)
    {
      strcpy(errstr, "Missing Parameters in SVSHOST");
      return 1;
    }
    if ( (u = finduser(parv[0])) )
    {
#ifdef MONSERV
      DelUserFromHashes(u);
#endif

      Change_Host(u, parv[1]);
#ifdef MONSERV

      AddUserToHashes(u);
#endif

      return 0;
    }
    snprintf(errstr, BUFSIZE, "Unkown user: %s", parv[0]);
    return 1;
  }
  strcpy(errstr, "No Origin in SVSHOST");
  return 1;
}

static int m_svsident(char *origin, int parc, char *parv[])
{
  struct User *u;

  if (origin)
  {
    if (parc < 2)
    {
      strcpy(errstr, "Missing Parameters in SVSDENT");
      return 1;
    }
    if ( (u = finduser(parv[0])) )
    {
#ifdef MONSERV
      DelUserFromHashes(u);
#endif

      Change_Ident(u, parv[1]);
#ifdef MONSERV

      AddUserToHashes(u);
#endif

      return 0;
    }
    snprintf(errstr, BUFSIZE, "Unkown user: %s", parv[0]);
    return 1;
  }
  strcpy(errstr, "No Origin in SVSIDENT");
  return 1;
}
#endif

#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
static int m_svsnick(char *origin, int parc, char *parv[])
{
  struct User *u;

  if (origin)
  {
#ifdef IRCD_HYBBG
    if (parc < 3)
#else

    if (parc < 2)
#endif

    {
      strcpy(errstr, "Missing Parameters in SVSNICK");
      return 1;
    }
    /* XXX Track stats on origin */
#ifdef IRCD_HYBBG
    if ( (u = finduser(parv[1])) )
#else

    if ( (u = finduser(parv[0])) )
#endif

    {
#ifdef IRCD_HYBBG
      Change_User(u, parv[2], 1);
#else

      Change_User(u, parv[1], 1);
#endif

      return 0;
    }
    snprintf(errstr, BUFSIZE, "Unknown user: %s", parv[0]);
    return 1;
  }
  strcpy(errstr, "No Origin in SVSNICK");
  return 1;
}

static int m_svsmode(char *origin, int parc, char *parv[])
{
  /* Old format:  0 = nick, 1 = mode, 2 = servicesID
  ** New format:  0 = nick, 1 = ts, 2 = mode, 3 = optional servicesID
  */
  char *modes;
  if (origin)
  {
    if (parc < 3)
    {
      strcpy(errstr, "Missing parameters in MODE");
      return 1;
    }
    /* SVSMODE has incompatabilities... */
    if ((parv[1][0] == '+') || (parv[1][0] == '-'))
      modes = parv[1];
    else
      modes = parv[2];
    /* XXX do channel modes if there are any */
    if ( !is_channel(parv[0]) )
      UserMode(finduser(parv[0]), modes);
    return 0;
  }

  strcpy(errstr, "SVSMODE without an origin");
  return 1;
}

static int m_svskill(char *origin, int parc, char *parv[])
{
  if (parc < 2)
  {
    strcpy(errstr, "No target in SVSKILL");
    return 1;
  }
  if (origin)
  {
    spy("SvsKill Request by %s for %s path: %s", origin, parv[0],
        (parv[2]) ? parv[2] : "NO REASON");
    sendto_dcc_flag(CF_SERVICES,
                    "SvsKill Request by %s for %s path: %s", origin,
                    parv[0], (parv[2]) ? parv[2] : "NO REASON");
    /* XXX - Track SVS in server structs */
    DelUser(parv[0]);
    if (!irccasecmp(parv[0], AdminServ.nick))
    {
#if defined(SPYMODE) || defined(ALERTMODE) || defined(DRONES_CHANNEL)
      ad_delayed = 1;
      eventAdd("check admin delayed", evh_CheckAdDelayed, NULL, DEFAULT_DELAY);
#endif

      init_admin();
#ifdef SPYMODE

      init_spy_join();
#endif
#ifdef ALERTMODE

      init_alert_join();
#endif
#ifdef DRONES_CHANNEL

      init_drone_join();
#endif

      HandleCollide();
#ifdef MONSERV

    }
    else if (!irccasecmp(parv[0], MonServ.nick))
    {
      init_monserv();
      HandleCollide();
#ifdef TRACK_CTCPVERSION

      mn_delayed = 1;
      eventAdd("check monserv delayed", evh_CheckMnDelayed, NULL,
               VERSION_DELAY);
#endif

#endif
#ifdef STATSERV

    }
    else if (!irccasecmp(parv[0], StatServ.nick))
    {
      init_statserv();
      init_join();
#ifdef PUBSTATS

      init_pubjoin();
#endif

      HandleCollide();
#endif
#ifdef SPLITSERV

    }
    else if (!irccasecmp(parv[0], SplitServ.nick))
    {
      init_splitserv();
      init_spjoin();
      HandleCollide();
#endif
#ifdef JUPESERV

    }
    else if (!irccasecmp(parv[0], JupeServ.nick))
    {
      init_jupeserv();
      HandleCollide();
#endif
#ifdef LISTSERV

    }
    else if (!irccasecmp(parv[0], ListServ.nick))
    {
      init_listserv();
      HandleCollide();
#endif

    }
    return 0;
  }

  strcpy(errstr, "SVSKILL without an origin");
  return 1;
}

/* XXX - Should I track SVSNOOP here at all? */

#ifndef IRCD_HYBSVS
static int m_burst(char *origin, int parc, char *parv[])
{
  if (parc < 1)
  {
    /* Start of burst, we don't care about that
    **
    ** but maybe we should?
    */ 
    return 0;
  }
  if (me.uplink != NULL)
  {
    if (strcasecmp(me.uplink->name, origin) == 0)
    {
      SetEOB();
      me.eob_time = time(NULL);
      if (IsNotifyEOB())
      {
        wallops(settings.name, "Burst from the uplink "
                "completed: %d seconds",
                me.eob_time - me.uplink->connected_since);
      }
      return 0;
    }
    strcpy(errstr, "BURST from a server other than the origin");
    return 1;
  }
  strcpy(errstr, "BURST received without being connected to a server");
  return 1;
}
#endif /* IRCD_HYBSVS */

#endif /* IRCD_BAHAMUT */

static void ircd_initialize(void)
{
#ifdef IRCD_BAHAMUT
  sts(":%s BURST", settings.name);
#endif
#ifdef STATSERV

  init_statserv();
#endif
#ifdef MONSERV

  init_monserv();
#endif
#ifdef SPLITSERV

  init_splitserv();
  init_spjoin();
#endif

  init_admin();
#ifdef JUPESERV

  init_jupeserv();
  FlushJupeList();
  LoadJupes();
#endif
#ifdef LISTSERV
  init_listserv();
#endif
#ifdef STATSERV

  init_join();
#ifdef PUBSTATS

  init_pubjoin();
#endif
#endif
#ifdef SPYMODE

  init_spy_join();
#endif
#ifdef ALERTMODE

  init_alert_join();
#endif
#ifdef DRONES_CHANNEL

  init_drone_join();
#endif
#if defined(IRCD_HYB7) || defined(IRCD_HYBSVS)

  sts(":%s EOB", settings.name);
#endif
#ifdef IRCD_BAHAMUT

  sts(":%s BURST 0", settings.name);
#endif
}

/* Send a message to a user from a user */
void notice(char *to, const char *from, char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE], buf2[BUFSIZE];
  char *cmd;

  if (is_channel(to))
    cmd = "PRIVMSG";
  else
    cmd = "NOTICE";

  va_start(ap, fmt);
  vsnprintf(buf2, sizeof(buf2), fmt, ap);
  snprintf(buf, BUFSIZE, ":%s %s %s :%s", from, cmd, to, buf2);
  sts("%s", buf);
  va_end(ap);
}

/* Read a help file and send it to the recipient via notice() */
void send_help_file(char *to, char *from, char *filename, char *query)
{
  FILE *fp = fopen(filename, "r");
  char buffer[BUFSIZE];

  if (!fp)
  {
    notice(to, from, "No help available for: %s", query);
    slog(DEFAULT_LOG, L_ERR, "No help available for %s, filename %s", query,
         filename);
    return ;
  }

  while (fgets(buffer, BUFSIZE, fp))
  {
    strip(buffer);
    if (buffer[0])
      notice(to, from, "%s", buffer);
    else
      notice(to, from, " ");
  }

  fclose(fp);
}

void wallops(char *from, char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZE], buf2[BUFSIZE];

  if (IsSilent())
    return ;

  va_start(ap, fmt);
  vsnprintf(buf2, sizeof(buf2), fmt, ap);
  snprintf(buf, BUFSIZE, ":%s WALLOPS :%s", from, buf2);
  sts("%s", buf);
  va_end(ap);
}

#ifdef STATSERV 
/* This function KILL's or optionally ignores a user for flooding services. */
/* This needs to be adjusted for SplitServ too! */
int flood(struct User *u)
{
  /* Our "opers" are immune to flood control */
  if (u->myuser)
    return 0;

  /* So are IRC opers (+o) */
  if (IsOper(u))
    return 0;

  /* Flood control is disabled by setting the flood lines to 0 */
  if (settings.flood)
  {
    time_t current = time(NULL);

    if (current - u->t_flood > settings.flood_time)
    {
      u->t_flood = current;
      u->flood = 0;
      u->flags &= ~UFLAG_FLOODSENT;
      return 0;
    }

    if (u->flood >= settings.flood)
    {
      if (!IsFlood())
      {
        sts(":%s KILL %s :%s!%s (Flooding Services.)",
            StatServ.nick, u->nick, StatServ.host,
            StatServ.nick);
        slog(DEFAULT_LOG, L_WARNING, "FLOODING (KILLED): %s!%s@%s", u->nick,
             u->username, u->hostname);
        DelUserU(u);
      }
      else
      {
        /* Only send the flood WALLOPS/GLOBOPS message
        ** once, or WE will be the ones flooding
        */
        if (!IsFloodSent(u))
        {
          wallops("%s is Flooding Services", u->nick);
          slog(DEFAULT_LOG, L_WARNING, "FLOODING (IGNORED): %s!%s@%s", u->nick,
               u->username, u->hostname);
        }
      }
      return 1;
    }
    else
    {
      u->flood++;
    }
  }
  return 0;
}
#endif

/* Send out QUIT messages as a courtesy... */
void SendQuit(struct User *u, int restart, char * comm)
{
#ifdef STATSERV
  sts(":%s QUIT :%s Request by %s: %s", StatServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef SPLITSERV

  sts(":%s QUIT :%s Request by %s: %s", SplitServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef JUPESERV

  sts(":%s QUIT :%s Request by %s: %s", JupeServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef LISTSERV

  sts(":%s QUIT :%s Request by %s: %s", ListServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
#endif
#ifdef MONSERV

  sts(":%s QUIT :%s Request by %s: %s", MonServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
#endif

  sts(":%s QUIT :%s Request by %s: %s", AdminServ.nick,
      (restart) ? "Restart" : "Shutdown", u->nick,
      (comm && comm[0]) ? comm : "No Reason");
}

/* This is how we answer STATS requests...
** This is very Hybrid-like at the moment.  It should not affect other
** ircd's.  If someone can refine this for other ircd's, I will patch it in
*/
static void ParseStats(struct User *u, char type)
{
  switch (type)
  {
    case 'C':
    case 'c':
      /* Our only "C and N" lines are to our uplink, show the
      ** real IP/address of our uplink to opers, but use the
      ** Hybrid6 method of IP hiding for non-opers.  This is
      ** moot if your ircd blocks remote STATS from non-opers
      */
      sts(":%s %d %s C %s A %s %d uplink", settings.name,
          RPL_STATSCLINE, u->nick,
#ifdef HIDE_IPS
          "*@127.0.0.1",
#else
          IsOper(u) ? settings.uplink : "*@127.0.0.1",
#endif
          me.uplink->name, settings.port);
      /* Removed N lines */
      break;
    case 'E':
    case 'e':
      /* Only show events to logged in users */
      if (u->myuser != NULL)
      {
        struct ev_enum *ev;
        const char *tmpe;

        tmpe = get_last_event_ran();
        if (tmpe != NULL)
          sts(":%s %d %s :*** Last event to run: %s", settings.name,
              RPL_STATSDEBUG, u->nick, tmpe);

        sts(":%s %d %s :*** Operation            Next Execution", settings.name,
            RPL_STATSDEBUG, u->nick);

        start_enumeration();
        ev = enumerate_events();
        while (ev != NULL)
        {
          sts(":%s %d %s :*** %-20s %-3d seconds", settings.name,
              RPL_STATSDEBUG, u->nick, ev->name, ev->next);
          ev = enumerate_events();
        }

        sts(":%s %d %s :*** Finished", settings.name, RPL_STATSDEBUG, u->nick);
      }
      break;
    case 'F':
    case 'f':
      {
        dlink_node *dl;
        struct Connection *conn;

        DLINK_FOREACH(dl, connection_list.head)
        {
          conn = (struct Connection *) dl->data;
          if (conn->flags & CONN_DEAD)
          {
            sts(":%s %d %s :Dying fd: %d name: %s",
                settings.name, RPL_STATSDEBUG,
                u->nick, conn->fd, conn->name);
          }
          else if (conn->flags & CONN_DCCIN)
          {
            if (conn->flags & CONN_CONNECTING)
              sts(":%s %d %s :Listening fd: %d name: %s",
                  settings.name, RPL_STATSDEBUG,
                  u->nick, conn->fd, conn->name);
            else
              sts(":%s %d %s :Inbound DCC fd: %d name: %s",
                  settings.name, RPL_STATSDEBUG,
                  u->nick, conn->fd, conn->name);
          }
          else if (conn->flags & CONN_DCCOUT)
          {
            if (conn->flags & CONN_CONNECTING)
              sts(":%s %d %s :Connecting fd: %d name: %s",
                  settings.name, RPL_STATSDEBUG,
                  u->nick, conn->fd, conn->name);
            else
              sts(":%s %d %s :Outbound DCC fd: %d name: %s",
                  settings.name, RPL_STATSDEBUG,
                  u->nick, conn->fd, conn->name);
          }
          else if (conn->flags & CONN_SERV)
          {
            sts(":%s %d %s :Server fd: %d",
                settings.name, RPL_STATSDEBUG,
                u->nick, conn->fd);
          }
          else
          {
            sts(":%s %d %s :Unknown fd: %d name: %s",
                settings.name, RPL_STATSDEBUG,
                u->nick, conn->fd, conn->name);
          }
        }
#if 0
        /* XXX - Renable these */
        if (main_log != NULL)
          sts(":%s %d %s :Log fd: %d name: %s", settings.name,
              RPL_STATSDEBUG, u->nick, fileno(main_log),
              LOGFILE);
#ifdef SPLITSERV

        if (split_log != NULL)
          sts(":%s %d %s :Log fd: %d name: %s", settings.name,
              RPL_STATSDEBUG, u->nick, fileno(split_log),
              SPLITLOG);
#endif
#ifdef IO_LOG

        if (io_log != NULL)
          sts(":%s %d %s :Log fd: %d name: %s", settings.name,
              RPL_STATSDEBUG, u->nick, fileno(io_log),
              IOLOGFILE);
#endif
#if defined(MONSERV) && defined(MON_DRONES)

        if (drone_log != NULL)
          sts(":%s %d %s :Log fd: %d name: %s", settings.name,
              RPL_STATSDEBUG, u->nick, fileno(drone_log),
              DRONELOG);
#endif

        if (spy_log != NULL)
          sts(":%s %d %s :Log fd: %d name: %s", settings.name,
              RPL_STATSDEBUG, u->nick, fileno(spy_log),
              SPYLOG);
#endif

        break;
      }
    case 'I':
    case 'i':
      /* This is purely cosmetic, show a closed server */
      sts(":%s %d %s I * * *@%s 0 nonopered", settings.name,
          RPL_STATSILINE, u->nick, settings.name);
      break;
    case 'H':
    case 'h':
      /* Our only hub should be our uplink.  This is cosmetic
      ** as well.
      */
      sts(":%s %d %s H * * %s", settings.name, RPL_STATSHLINE,
          u->nick, me.uplink->name);
      break;
    case 'M':
    case 'm':
      {
        int i;

        /* Show the known commands and their usage counters */
        for (i = 0; msg_table[i].handler; i++)
        {
          /* lie about the byte count */
          sts(":%s %d %s %s %ld 0", settings.name, RPL_STATSCOMMANDS,
              u->nick, msg_table[i].name, msg_table[i].counter);
        }
        break;
      }
    case 'O':
    case 'o':
      {
        dlink_node *dl;
        struct MyUser *myu;
        /* This STATS is actually very useful.  This shows a list of
        ** our users (all who can use LOGIN, the equivalent of opers)
        */
        DLINK_FOREACH(dl, myuhead.head)
        {
          myu = (struct MyUser *) dl->data;
          sts(":%s %d %s O *@* * %s 0 opered", settings.name,
              RPL_STATSOLINE, u->nick, myu->username);
        }
        break;
      }
      /* P is ports, we don't listen (yet) so ignore it */
    case 'p':
      {
        dlink_node *dl;
        struct MyUser *myu;
        int count = 0;

        /* This STATS is also relatively useful.  It does the Hybrid
        ** STATS p - show all online opers.  All users who have used
        ** LOGIN will be shown as Sentinel opers.
        ** NOTE:  We now DO print out the current user's nick.
        */
        DLINK_FOREACH(dl, myuhead.head)
        {
          myu = (struct MyUser *) dl->data;
          if (IsOperOn(myu))
          {
            sts(":%s %d %s :[O] %s", settings.name, RPL_STATSDEBUG,
                u->nick, myu->u->nick);
            count++;
          }
          /* They CAN be online in both places...show both */
          if (IsOperDCC(myu))
          {
            sts(":%s %d %s :[D] %s", settings.name, RPL_STATSDEBUG,
                u->nick, myu->conn->name);
            count++;
          }
        }
        sts(":%s %d %s :%d OPER%s", settings.name, RPL_STATSDEBUG, u->nick,
            count, (count == 1) ? "" : "s");
        break;
      }
      /* U is U: lines/shared blocks, ignore */
    case 'u':
      /* Send our uptime in the usual fasion */
      sts(":%s %d %s :Statistics up %s", settings.name,
          RPL_STATSUPTIME, u->nick,
          timediff(time(NULL) - me.t_start));
      break;
    case 'T':
    case 't':
      /* For now, just show the byte counters */
      sts(":%s %d %s :Bytes in: %ld, Bytes out: %ld", settings.name,
          RPL_STATSDEBUG, u->nick, netinfo.inb, netinfo.outb);
      break;
    case 'Y':
    case 'y':
      /* Purely cosmetic with fake values...EXCEPT for the
      ** reconnect time in the server Y line
      */
      sts(":%s %d %s Y nonopered 0 0 2 0", settings.name,
          RPL_STATSYLINE, u->nick);
      sts(":%s %d %s Y uplink 0 %d 1 0", settings.name,
          RPL_STATSYLINE, u->nick, settings.r_time);
      sts(":%s %d %s Y opered 0 0 1 0", settings.name,
          RPL_STATSYLINE, u->nick);
      break;
    default:
      break;
  }
  sts(":%s %d %s %c :End of STATS report", settings.name,
      RPL_ENDOFSTATS, u->nick, type);
}

/* Report TRACE in the Hybrid style */
static void Trace(struct User *u)
{
#ifdef JUPESERV
  dlink_node *dl;
  struct Jupe *j;
#endif

  if (IsOper(u))
  {
#ifdef STATSERV
    sts(":%s %d %s %s %s %s[%s@%s] (127.0.0.1) 0 0",
        settings.name,
        IsStSetOper() ? RPL_TRACEOPERATOR : RPL_TRACEUSER,
        u->nick,
        IsStSetOper() ? "Oper" : "User",
        IsStSetOper() ? "opered" : "nonopered", StatServ.nick,
        StatServ.user, StatServ.host);
#endif
#ifdef SPLITSERV

    sts(":%s %d %s User nonopered %s[%s@%s] (127.0.0.1) 0 0",
        settings.name, RPL_TRACEUSER, u->nick, SplitServ.nick,
        SplitServ.user, SplitServ.host);
#endif

    sts(":%s %d %s Oper opered %s[%s@%s] (127.0.0.1) 0 0",
        settings.name, RPL_TRACEOPERATOR, u->nick, AdminServ.nick,
        AdminServ.user, AdminServ.host);
#ifdef MONSERV

    sts(":%s %d %s Oper opered %s[%s@%s] (127.0.0.1) 0 0",
        settings.name, RPL_TRACEOPERATOR, u->nick, MonServ.nick,
        MonServ.user, MonServ.host);
#endif
#ifdef LISTSERV

    sts(":%s %d %s User nonopered %s[%s@%s] (127.0.0.1) 0 0",
        settings.name, RPL_TRACEUSER, u->nick, ListServ.nick,
        ListServ.user, ListServ.host);
#endif
#ifdef JUPESERV

    sts(":%s %d %s Oper opered %s[%s@%s] (127.0.0.1) 0 0",
        settings.name, RPL_TRACEOPERATOR, u->nick, JupeServ.nick,
        JupeServ.user, JupeServ.host);
    DLINK_FOREACH(dl, server_jupes.head)
    {
      j = (struct Jupe *) dl->data;
      sts(":%s %d %s Serv jupedserver 0S 0C %s %s!*@%ld 0",
          settings.name, RPL_TRACESERVER, u->nick, j->jupe, j->oper, j->ts);
    }
    DLINK_FOREACH(dl, nick_jupes.head)
    {
      j = (struct Jupe *) dl->data;
      sts(":%s %d %s User jupednick %s[jupe@juped.nick] (127.0.0.1) 0 0",
          settings.name, RPL_TRACEUSER, u->nick, j->jupe);
    }
#endif

    sts(":%s %d %s Serv uplink %dS %ldC %s AutoConn.!*@%s 0",
        settings.name, RPL_TRACESERVER, u->nick, netinfo.maxservers,
        netinfo.maxusers, me.uplink->name, settings.name);
    sts(":%s %d %s %s :End of TRACE", settings.name, RPL_ENDOFTRACE, u->nick,
        settings.name);
  }
}

#ifdef SUPPORT_MSG_OPERS
static void MsgOpers(char *nick, char *message)
{
  struct MyUser* myu = NULL;
  dlink_node *dl;

  DLINK_FOREACH(dl, myuhead.head)
  {
    myu = (struct MyUser *) dl->data;

    if (IsOperOn(myu))
      sts(":%s NOTICE %s :To opers: From %s: %s",
          settings.name, myu->u->nick, nick, message);
    if (IsOperDCC(myu))
      sendto_dcc(myu->conn->fd, "To opers: From %s: %s",
                 nick, message);
  }
}
#endif

static void CollideDie()
{
  wallops(settings.name, "Collision Flood!  Gracefully Exiting");
  slog(DEFAULT_LOG, L_CRIT, "Collision Flood!  Gracefully Exiting");
#ifdef STATSERV

  SendQuit(StatServ.u, 0, "Collision Flood");
#endif
#ifdef SPLITSERV

  SendQuit(SplitServ.u, 0, "Collision Flood");
#endif
#ifdef JUPESERV

  SendQuit(JupeServ.u, 0, "Collision Flood");
#endif
#ifdef LISTSERV

  SendQuit(ListServ.u, 0, "Collision Flood");
#endif
#ifdef MONSERV

  SendQuit(MonServ.u, 0, "Collision Flood");
#endif

  SendQuit(AdminServ.u, 0, "Collision Flood");
  sleep(1);
  close(servsock);
  close_dccs();
  SaveStats();
  close_logfiles();
  exit(0);
}

static void HandleCollide()
{
#ifdef DIE_ON_MASS_COLLIDE
  netinfo.collides++;
  if (netinfo.collides > MAXCOLLISIONS)
  {
    CollideDie();
  }
#endif
}
